from abc import ABC, abstractmethod
from typing import List

import numpy as np
import numpy.typing as npt

from sziszapangma.core.transformer.embedding_transformer import EmbeddingTransformer
from sziszapangma.model.model import Word


class DistanceCalculator(ABC):
    @abstractmethod
    def calculate_distance_matrix(
        self, reference: List[Word], hypothesis: List[Word]
    ) -> npt.NDArray[np.float64]:
        pass

    @abstractmethod
    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        pass


class BinaryDistanceCalculator(DistanceCalculator):
    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        return 0 if word1["text"] == word2["text"] else 1

    def calculate_distance_matrix(
        self, reference: List[Word], hypothesis: List[Word]
    ) -> npt.NDArray[np.float64]:
        return np.array(
            [
                [
                    self.calculate_distance_for_words(reference_word, hypothesis_word)
                    for hypothesis_word in hypothesis
                ]
                for reference_word in reference
            ]
        )


class CosineDistanceCalculator(DistanceCalculator):
    _embedding_transformer: EmbeddingTransformer

    def __init__(self, embedding_transformer: EmbeddingTransformer):
        self._embedding_transformer = embedding_transformer

    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        return self.cosine_distance_between_words_embeddings(
            self._embedding_transformer.get_embedding(word1["text"]),
            self._embedding_transformer.get_embedding(word2["text"]),
        )

    @staticmethod
    def cosine_distance_between_words_embeddings(
        word1_embedding: npt.NDArray[np.float64], word2_embedding: npt.NDArray[np.float64]
    ) -> float:
        a = word1_embedding
        b = word2_embedding
        if a.shape != b.shape:
            raise RuntimeError("array {} shape not match {}".format(a.shape, b.shape))
        if a.ndim == 1:
            a_norm = np.linalg.norm(a)
            b_norm = np.linalg.norm(b)
        elif a.ndim == 2:
            a_norm = np.linalg.norm(a, axis=1, keepdims=True)
            b_norm = np.linalg.norm(b, axis=1, keepdims=True)
        else:
            raise RuntimeError("array dimensions {} not right".format(a.ndim))
        similarity = np.dot(a, b.T) / (a_norm * b_norm)
        dist = 1.0 - similarity
        # return float(dist)
        float_dist = float(dist)
        if abs(float_dist) < 0.000001:
            return 0.0
        else:
            return float_dist

    def calculate_distance_matrix(
        self, reference: List[Word], hypothesis: List[Word]
    ) -> npt.NDArray[np.float64]:
        embeddings_dict = self._embedding_transformer.get_embeddings(
            list(set(it["text"] for it in (reference + hypothesis)))
        )
        return np.array(
            [
                [
                    self.cosine_distance_between_words_embeddings(
                        embeddings_dict[reference_word["text"]],
                        embeddings_dict[hypothesis_word["text"]],
                    )
                    for hypothesis_word in hypothesis
                ]
                for reference_word in reference
            ]
        )
